#include <mrpt/poses.h>
#include <mrpt/math.h>
#include <math.h>
#include<vector>
#include<iostream>
#include <mrpt/base/include/mrpt/math/CQuaternion.h>
#include <mrpt/base/include/mrpt/poses/CPose3D.h>
#include <mrpt/base/include/mrpt/poses/CPoint3D.h>
#include <mrpt/utils.h>
#include "Parser.h"

using namespace std;
using namespace mrpt::poses;
using namespace mrpt::math;
using namespace mrpt::utils;

struct data_entry{
    size_t timestamp;
    CPose3D data;
};

vector<data_entry> ground_truth;
vector<data_entry> slam_output_orb;
vector<data_entry> slam_output_icp;
vector<data_entry> slam_output_zed;


bool prepare_for_ekf;

/// Loads dataset on specified path
/// \param path to data
/// \param dataset to store loaded data
void load_data(string path,vector<data_entry> & dataset);

/// Transform data into a correct coordinate system
/// \param dataset
/// \param offset of the first measurement
/// \param cam transformation between coordinate systems
void offset_data(vector<data_entry> & dataset,CPose3D offset,CPose3D cam);

/// Computes angular error based on unit quaternoin metric
/// \param ground_pose ground truth
/// \param slam_pose slam output
/// \return angular distance in degrees
double compute_ang_error(CPose3D ground_pose ,CPose3D slam_pose);

/// Computes translational error based on Euclidean metric
/// \param ground_pose ground truth
/// \param slam_pose slam output
/// \return distance in metres
double compute_trans_error(CPose3D ground_pose,CPose3D slam_pose);

/// Computes an overall trajectory of the whole dataset
/// \param dataset to be evaluated
/// \return trajectory in metres
double compute_trajectory(vector<data_entry> &dataset);

/// Computes Euclidean distance of two poses
/// \param a pose a
/// \param b pose b
/// \return distance in metres
double compute_distance(CPose3D a,CPose3D b);

/// Saves data as 2D coordinates [meters] with heading [radians]
/// \param dataset
/// \param dest
void save_for_matlab(vector<data_entry> &dataset, string dest);
/// Prints statistics
/// \param dataset1
/// \param ground_truth
/// \param label
void print_statistics(vector<data_entry> &dataset1,vector<data_entry> &ground_truth,string label);
/// Pairs ground truth with corresponding slam output
/// \param dataset1
/// \param ground_truth
/// \return paired ground truth
vector<data_entry> pair_ground_truth(vector<data_entry> &dataset1,vector<data_entry> &ground_truth);

/// Saves instanoues translational and angular error for plotting in matlab
/// \param dataset
/// \param ground_truth
/// \param dest
void save_error_matlab(vector<data_entry> &dataset,vector<data_entry> &ground_truth, string dest);

int main(int argc, char** argv) {

    ///Parse commandline
    Parser parser;
    if(parser.process_arguments(argc,argv)) return EXIT_SUCCESS;
    string path = parser.init.path;
    prepare_for_ekf = parser.init.use_ekf;

    ///Strings for data operation
    string ground = path+"ground_truth.txt";
    string ground_matlab = path+"ground_truth_matlab.csv";
    
    string slam_orb = path+"ORB_OUTPUT.csv";
    string slam_orb_matlab = path+"ORB_OUTPUT_MATLAB.csv";
    
    string slam_icp = path+"ICP_OUTPUT.csv";
    string slam_icp_matlab = path+"ICP_OUTPUT_MATLAB.csv";
    
    string slam_zed = path+"ZED_slam_output.csv";
    string slam_zed_matlab = path+"ZED_OUTPUT_MATLAB.csv";

    string odom = path+"odometry.txt";
    string odom_matlab = path+"odometry_matlab.csv";

    ///Loads data from files
    load_data(ground,ground_truth);
    load_data(slam_orb,slam_output_orb);
    load_data(slam_icp,slam_output_icp);
    load_data(slam_zed,slam_output_zed);

    ///Coordinates transform
    CPose3D cam1(0,0,0,DEG2RAD(90),0,0);
    ///orb-slam
    CPose3D cam2(0,0,0,0,DEG2RAD(90.0),0);
    CPose3D cam3(0,0,0,DEG2RAD(-90.0),0,0);
    CPose3D cam =  -cam2 -cam3;

    ///Offset data
    offset_data(ground_truth,ground_truth.front().data,cam1);
    offset_data(slam_output_orb,slam_output_orb.front().data,cam);

    ///Statistics
    print_statistics(slam_output_icp,ground_truth,"ICP");
    print_statistics(slam_output_orb,ground_truth,"ORB");
    print_statistics(slam_output_zed,ground_truth,"ZED");

    ///Pair ground truth
    vector<data_entry> paired_ground = pair_ground_truth(slam_output_zed,ground_truth);

    if(prepare_for_ekf) {
        ///Save for matlab for EKF
        save_for_matlab(slam_output_icp, slam_icp_matlab);
        save_for_matlab(slam_output_zed, slam_zed_matlab);
        save_for_matlab(slam_output_orb, slam_orb_matlab);
        save_for_matlab(ground_truth, ground_matlab);
    } else{
        ///Save for matlab for plotting
        save_error_matlab(slam_output_orb,paired_ground, slam_orb_matlab);
        save_error_matlab(slam_output_zed,paired_ground, slam_zed_matlab);
        save_error_matlab(slam_output_icp,paired_ground, slam_icp_matlab);
    }
    return EXIT_SUCCESS;
}

void load_data(string path,vector<data_entry> &dataset){
    ifstream infile(path);
    double data[8];
    int index = 0;
    if (infile.is_open()) {
        string str;
            while (infile >> str){
                str.erase(std::remove(str.begin(), str.end(), ','), str.end());
                data[index]=atof(str.c_str());
                index++;              
                if(index==8){
                    double norm = 0.0;
                    for (int i = 4; i < 8; i++) {
                        norm += pow(data[i],2);
                    }
                    norm = sqrt(norm);
                    CQuaternion<double> quat(data[7]/norm,data[4]/norm,data[5]/norm,data[6]/norm);
                    CPose3D pose(quat,data[1],data[2],data[3]);
                    size_t stamp = (size_t) data[0];
                    data_entry temp;
                    temp.timestamp = stamp;
                    temp.data = pose;
                    dataset.push_back(temp);                       
                    index = index % 8;                   
                }
            }
    }
    infile.close();
}
void offset_data(vector<data_entry> &dataset,CPose3D offset,CPose3D cam){
    for (size_t i = 0; i < dataset.size(); i++) {   
        dataset.at(i).data = (dataset.at(i).data + cam)- ( offset  + cam);

    }
}
double compute_ang_error(CPose3D ground_pose ,CPose3D slam_pose){

    CQuaternion<double> ground;
    CQuaternion<double> slam;
    ground_pose.getAsQuaternion(ground);
    slam_pose.getAsQuaternion(slam);
    double temp = ground.dot(slam);
    temp = 2*acos(abs(temp));
    return RAD2DEG(temp);

}
double compute_trans_error(CPose3D ground_pose,CPose3D slam_pose){
    double distance = ground_pose.sqrDistanceTo(slam_pose);
    return sqrt(distance);

}
double compute_trajectory(vector<data_entry> &dataset){
    double trajectory = 0.0;

    for (size_t i = 1; i < dataset.size(); i++) {       
        trajectory+=compute_distance(dataset.at(i).data,dataset.at(i-1).data);
    }
    return trajectory;
}
double compute_distance(CPose3D a,CPose3D b){
    double trajectory = pow(a.x()-b.x(),2)+pow(a.y()-b.y(),2)+pow(a.z()-b.z(),2);
    return sqrt(trajectory);
}

void save_for_matlab(vector<data_entry> &dataset, string dest){
    ofstream outputFile;
    outputFile.open(dest.c_str());
    for (auto elem : dataset) {
        outputFile << elem.timestamp<< ", " << elem.data.x()<< ", " << elem.data.y() << ", " << elem.data.yaw() << endl;
    }
    outputFile.close();
}
void save_error_matlab(vector<data_entry> &dataset,vector<data_entry> &ground_truth, string dest){
    ofstream outputFile;
    outputFile.open(dest.c_str());
    for (int i = 0; i < dataset.size(); ++i) {
        outputFile << dataset.at(i).timestamp << ", " << compute_trans_error(dataset.at(i).data,ground_truth.at(i).data)
        << ", " << compute_ang_error(dataset.at(i).data,ground_truth.at(i).data) << ", "  << endl;

    }
    outputFile.close();
}

void print_statistics(vector<data_entry> &dataset1,vector<data_entry> &ground_truth, string label){
    cout << "***************************************************************" << endl;
    cout << label<< endl;
    cout << "***************************************************************" << endl;
    cout << "Ground truth: " << ground_truth.back().data.getAsVectorVal().transpose() << endl;
    cout << label <<" SLAM: " << dataset1.back().data.getAsVectorVal().transpose() << endl;
    cout << "***************************************************************" << endl;
    cout << "Trans error: " << compute_trans_error(ground_truth.back().data,dataset1.back().data)<< " metres." << endl;
    cout << "Angular error: " << compute_ang_error(ground_truth.back().data,dataset1.back().data)<< " degress." << endl;
    cout << "Ground truth trajectory: " << compute_trajectory(ground_truth) << " metres."<< endl;
    cout << label<< " SLAM trajectory: " << compute_trajectory(dataset1) << " metres."<< endl;
    cout << "***************************************************************" << endl;
    cout << endl;
}
vector<data_entry> pair_ground_truth(vector<data_entry> &dataset,vector<data_entry> &ground_truth){
    vector<data_entry> ground_truth_paired;
    for (size_t i = 0; i < dataset.size(); ++i) {
        for (size_t j = 0; j <ground_truth.size() ; ++j) {
            if(dataset.at(i).timestamp <= ground_truth.at(j).timestamp){
                ground_truth_paired.push_back(ground_truth.at(j));
                break;
            }
            else if(j == ground_truth.size()-1){
                ground_truth_paired.push_back(ground_truth.at(j));
            }
        }
    }
    return ground_truth_paired;
}